#include "eeprom.h"
/* 
EEPROM Emulator for uPSD34xx		Copyright: (c) Tim Jackson, April 2007

This code is suppiied as-is and the author makes no guarantee that it is free from errors 
or that it is suitable for any particular purpose; and will accept no libaility for any 
losses consequential upon its use.

You are free to use or modify this code in your own applications at your own risk.

file eeprom.c

Compile with Raisonance compiler and Large memory model

In these procedures there are tens-of-microsecond waits to allow EEPROM to execute.
These are implemented by for loops which are guaranteed to take at least 10 cycles
after optimising, and so should take at least one microsecond per iteration.
Excess time is unimportant - they only ever complete in event of some sort of error, but
it is important that they do eventually time out and don't hang the program.
*/

volatile xdata at PSD_REG_ADDR PSD_REGS EE_PSD_reg;       // Access PSD regs at CSIOP addr. Sadly can't make this static
typedef xdata unsigned char *FLASHPTR;
struct PAGEHDR{
	unsigned char hdrlen;
	unsigned char recordsize;
	char erase; // -1 or 0 if none, pageno + 0x80 to erase
	char ident;
	char user[1];
};
xdata at EEPROMBASE union {
	unsigned char volatile EEPROM[EEPagesize];
	struct PAGEHDR pagehdr; 
} ROM;
static data FLASHPTR next=0;
static char nextlevel=-1, erase=-1, activepage=-1;
static generic void *eebuf, *flashbuf;
static unsigned char eesize, flashsize;

// Given a node and its level, return a pointer to the current data record
// Returns next=0 zero if page full
static xdata void * GetAddr ( void) {
	data unsigned char level, node, levelmask;
	int bigblock, smallblock;
	long superblock;
	bit bytechange;
	unsigned char recordsize = ROM.pagehdr.recordsize;
	FLASHPTR addr;
	
   addr = (FLASHPTR)(ROM.EEPROM + ROM.pagehdr.hdrlen);
   level = MAXLEVEL;
	smallblock = recordsize * 8 + 1;       // Blocksize for low-byte bits
	if( level > 1) {
		bigblock = ( smallblock * 4) + 1;   // Blocksize for 2nd-byte bits
		if( level > 9) {
			superblock = (long)bigblock * 256 + 1; // Blocksize for 3rd-byte bits
			levelmask = 1 << (level - 9);
		} else levelmask = 1 << (level - 2);
	} else levelmask = 0x80;
	node = *addr;
	next = 0;	// write-pointer returned is null if page is full.
	do {	// Traverse tree down to level 1
		bytechange = (level == 2) || (level == 10);
		if( (node & levelmask) == 0) {	// This block is full, move to child node
			if( level > 10) addr += superblock << (level - 11);
			else if( level > 2) addr += bigblock << (level - 3);
			else addr += smallblock << (level-1);
			node = *addr;
		} else {
			next = addr;      		// else bring write-pointer here
			nextlevel = level;
		} 
		if( bytechange) {				// move to lower node byte
			node = *(++addr); 
			levelmask = 0x80;
		}
		else levelmask >>= 1;
	} while( --level > 0);
// This node should have its high bit set.  If not the index is corrupted
// (and the while loop below would hang)
	if( (node & 0x80) == 0) return 0;
	if( node & 0x20) {	// There is at least one unused record in this block
		next = addr;		// so bring write-pointer to it  
		nextlevel = 0;	
	}
	addr++;									// Move onto first data record of block
	while( (node & 1) == 0) {	      // Process the bottom 7 bits to find the current record
		addr += recordsize;
		node >>= 1;
	}
	return (xdata void *)addr;
// "addr" now points to the current record
//	"next" now points to the next node to have a bit programmed
// High or low byte, as appropriate.
// "nextlevel" points to the bit within that node 
// (0 means bits 0-5, 1,2 means 6,7, 3-10 means 0-7 (points to an upper byte))
}

//Get a pointer to the next free data record in ROM
// This could be done in response to a power fail interrupt, 
// followed by writing the record then pseudo-erasing its predecessor
static xdata void * GetEEpromFree( void) {
	data unsigned char recordsize, node;
	unsigned int block;
	FLASHPTR addr;
	
// Scan the tree to get pointers to next node if necessary
	if( nextlevel < 0) GetAddr();
	if( next == 0) return 0;
	recordsize = ROM.pagehdr.recordsize;
	addr = next;   
// If we came in with a nonzero level then we apply an offset as if the relevant
// tree node were zero. That should fetch a node that is all ones so we can go
// straight to level zero. So check that it is and return its first data record
	if( nextlevel > 0) {
		block = recordsize * 8 + 1;
		if( nextlevel > 10) addr += ( ( block * 1024L) + 257L) << ( nextlevel - 11);
		else if( nextlevel > 2) addr += ( ( block * 4) + 1) << ( nextlevel - 3);
		else addr += block << ( nextlevel - 1);
		if( nextlevel > 1) addr++;		// Move to bitmap byte		
		if( nextlevel > 9) addr++;	
		if( *addr != 0xFF) return 0;	// Hit something dirty
		return ++addr;						// move to first record;
	}	
// If we came in at level zero then the block can't be empty, so if all bits set it means
// record zero is current and record one is free, etc. 
// Bit 5 clear means record 6 is current and record 7 is free.
	node = *addr++;  // if level zero, then get the bitmap and move over it			
	if( node & '\xC0' != '\xC0') return 0;
// If we get here without bits 6-7 set it is end of index or corrupt
// (and would make an endless loop)
	while( (node & '\x01') == 0) {	// find the current record
		addr+= recordsize;
		node >>= 1;
	}
	return addr + recordsize;  // go to the next record
}

enum FlashCmd { WRITE, BYPASS, ERASE};
static void EEpromCmd( enum FlashCmd command){
	data union {
		unsigned int hex;
		volatile FLASHPTR ptr;
	}	flashA, flash5;
	
// Unlock Flash
	flashA.ptr = ROM.EEPROM;
	flashA.hex &= 0xF000;
	flash5.hex = flashA.hex;
	flash5.hex |= 0x555;
	flashA.hex |= 0xAAA;
	*flash5.ptr = 0xAA;	//Write Unlock Flash sequence
	*flashA.ptr = 0x55;
	switch( command) {
	case WRITE: 
		*flash5.ptr = 0xA0;
		break;
	case BYPASS:
		*flash5.ptr = 0x20;
		break;
	case ERASE:	// Erase selected sector only
		*flash5.ptr = 0x80;	//Write sector Erase Flash sequence
		*flash5.ptr = 0xAA;
		*flashA.ptr = 0x55;
		*flashA.ptr = 0x30;
// Caller must stay on this page until erase completes, and be aware that it is in erase mode
		break;
	}
}	

// Reset the Flash system after an error
static void EEReset( void){
	data char c1, c2, i;

	*ROM.EEPROM = 0xF0;		// reset flash system
	c1 = *ROM.EEPROM;
	for( i=0; i<25; i++){// Wait up to 25uS for data mode
		c2 = *ROM.EEPROM;
	 	if( c1 == c2) break;  
	 	c1 = c2;
	}
}

// Program a flash byte and wait for it to be done
static bit EEProg( volatile FLASHPTR addr, unsigned char c){
	data unsigned char c1, i;

	*addr = c;
	for( i=0; i<150; i++) {			// Wait up to 150us for programming
		c1 = *addr;						// Checking for top bit inversion
		if( c1 == c) return TRUE;	// byte written
		if( c1 & 0x20) return FALSE;// Write error
	}
	return FALSE;						// Timeout
}

// Write 'len' bytes from 'source' to ROM.EEPROM at 'addr': ie header or record data
static bit WriteEEprom( void * source, unsigned int len, volatile FLASHPTR addr) {
	unsigned generic char *s = (unsigned char *)source; // (for compiler bug)
	data unsigned char n;
	
	EEpromCmd( BYPASS);					// Enable multiple write
	for( n=0; n<len; n++) { 			// Write data	
		*addr = 0xA0;					// Command write
		if( !EEProg( addr++, *s++)){		// byte written
			EEReset();
			return FALSE;	
		}
	}
	*addr = 0x90;						// Command end bypass
	*addr = 0;
	return TRUE;
}

// Write a single byte to ROM.EEPROM
static bit PutEEprom( unsigned char source, volatile FLASHPTR addr) {
	
	EEpromCmd( WRITE); 		// Enable single write
	if( EEProg( addr, source)) return TRUE;
	EEReset();
	return FALSE;	
}

// Write a record 's' to ROM.EEPROM at 'addr'
static bit WriteEEpromData( void * s, FLASHPTR addr) {
	int i, len;
	
	len = ROM.pagehdr.recordsize;	// Make sure called func doesn't evaluate this during write mode
	for( i=0; i<len; i++){		// Don't rely on writing ones to zeroes
		if(addr[i] != 0xFF) return FALSE;  // field not clean: go switch pages
	}
	return WriteEEprom( s, len, addr);
}

// Release page register for other uses after erase completes.
// DO NOT put in interrupt task - it can't run inside EESave
// Returns true if still erasing 
// If so, caller must use suspend() before changing pages.
bit EECheck( void){
	unsigned char c1,c2;
	c1 = *ROM.EEPROM;
	c2 = *ROM.EEPROM;
	if( (c1^c2) & 0x40) {			// Flash is in command mode
		if( (c2 & 0x20) == 0) return TRUE;      // Erase in progress.
// Flash is still busy but has set error flag
// This means that the 34xx was unable to erase	the sector.  Death of hardware.
// We may be able to carry on for a while, this sector will skip or run short, 
// so just reset the flag & call it erased.  
// "All sectors corrupt" will be detected elsewhere.
		EEReset();
	} 
// Flash is (now) in read mode	
	if( (erase >= 0) && (activepage >=0) ) { 	      // Was erasing
		EE_PSD_reg.PAGE = activepage;      // Clear the erase-pending marker
		PutEEprom( 0, (xdata unsigned char*)&ROM.pagehdr.erase);	     
		erase = -1;	
	}
	return FALSE;				// Erase completed
}

// Interrupt erase and reselect active page
// (Should only happen on a save, but could happen on restore if done more than once)
static void suspend( void){
	if( EECheck()){					// Erase in progress?
		unsigned char c1, c2, i;
		
		*ROM.EEPROM = 0xB0;			// Suspend command;
		erase = EE_PSD_reg.PAGE;  	// It already should be, but don't take chances
		c2 = *ROM.EEPROM; 		
		for( i=0; i<15; i++) {  // Wait up to 15uS for data access mode
			c1 = *ROM.EEPROM; 		
			if( c1 == c2) break;
			c2 = c1;
		}
		if( activepage >= 0) EE_PSD_reg.PAGE = activepage;
	}
}

// Resume any suspended erase.  Returns true if no erase, ie free to switch pages
static bit resume( void){
	if( erase < 0) return TRUE;
	EE_PSD_reg.PAGE = erase;		// Re-select page being erased	
	*ROM.EEPROM = 0x30;          // and resume it.
	return FALSE;
}
 
// Initialise next available blank page 
// Return true on success
static bit EEInit() {
	struct PAGEHDR hdr;
	char p = activepage, i;
	bit writeok;
	
	for( i=0; i<EEPages; i++){    	// Search for a clean page starting after the active page
		if( ++p >= EEPages) p = 0;
		EE_PSD_reg.PAGE = p;
		if( ROM.pagehdr.ident == (char)-1) {	// Hopefully, a clean page)
			hdr.hdrlen = sizeof( hdr) + flashsize - 1;	// hdr declaration has dummy flash record of one byte.
			hdr.recordsize = eesize;
			hdr.erase = -1;	// Don't write the erase flag until the new data has been written
			hdr.ident = 'E';
			writeok = WriteEEprom( &hdr, sizeof( hdr)-1, ROM.EEPROM);			 // header without dummy flash byte
			if( (flashsize > 0) && !WriteEEprom( flashbuf, flashsize, ROM.EEPROM + (sizeof( hdr)-1))) writeok = FALSE;
			if( writeok) { 
				PutEEprom( 0, ROM.EEPROM + (EEPagesize - 1));  // Put a stopper at the end
				activepage = p;
				nextlevel = 0;
				next = ROM.EEPROM + ROM.pagehdr.hdrlen + (MAXLEVEL+6) / 8; // Address first node low byte
				return TRUE;			
			}
		}
	}
	if( activepage >= 0) EE_PSD_reg.PAGE = activepage;
	return FALSE;	// Failed to find a blank page - probably hardware failure (or PSD config error)
}

// Find latest record and read into RAM (eg. on power-up) 
// On coldstart create EEProm using 'init' as initial record value
// Returns true on coldstart, false on restore
// In a failed coldstart, the init string is not copied to the image.
bit EERestore( void *image, unsigned char imgsize, void *flash, unsigned char usersize){
	FLASHPTR idata source;
	idata int i;
// cimage is there because the compiler gives a spurious warning on incrementing a cast void pointer
	generic unsigned char * idata cimage;
	
	suspend();
	activepage = -1;		// default if we can't find anything
	flashbuf = flash;
	eebuf = image;
	eesize = imgsize;
	flashsize = usersize;
	for( i=0; i < EEPages; i++) {
		EE_PSD_reg.PAGE = i;
		if( ROM.pagehdr.ident == 'E') {
			activepage = i;
			source = GetAddr();
			if( source) {			// Found a valid-looking EEPROM page
				cimage = image;          				// Copy ROM.EEPROM record to RAM
				if( eesize > ROM.pagehdr.recordsize) eesize = ROM.pagehdr.recordsize; 
				for( i = 0; i<eesize; i++) *cimage++ = *source++;
				cimage = flash;          				// Copy Flash record to RAM
				source = (FLASHPTR)ROM.pagehdr.user;
				flashsize = ROM.pagehdr.hdrlen - ( source - ROM.EEPROM);	// Size of flash record in ROM
				if( flashsize > usersize) flashsize = usersize;        // protect against buffer overrun
				for( i = 0; i<flashsize; i++) *cimage++ = *source++;				
				if( resume()){						// if nothing to resume check for a full page
					erase = ROM.pagehdr.erase;		// If a page needs erasing, 0x80 is set and it isn't 0xFF
					if( ( erase != -1) && (erase & 0x80)) {
						erase &= 0x7F;
						EE_PSD_reg.PAGE = erase;	// Select the old page and erase it
						EEpromCmd( ERASE); 
					} else { // Nothing queued for erasing.  Check for orphan sectors
						for( erase=0; erase<EEPages; erase++){
							if( erase != activepage) {
								EE_PSD_reg.PAGE = erase;
								if( ROM.pagehdr.ident != 0xFF) {
									EEpromCmd( ERASE);				// Start erasing an orphan page
									return FALSE;
								}
							}
						}
						erase = -1;	// No orphans found
					}
				} 
/* 
It would be a good idea to check now that there is a clear landing zone for next write.
That means that after writing the last record on power down, we will immedaitely rewrite it
as the first record of a new page on power up.  This seeems wasteful, but it halves the worst 
case time spent in power fail by never writing a new header then.
*/
				return FALSE;
			}
		}
	} 
	EEInit( );	// No valid EEprom page found - do coldstart, assume it worked
// We don't really need to write a default record, but there's no way of marking the database as empty
	WriteEEpromData( image, next+1);       // Copy default value to first record
	resume();  		// shouldn't happen, but...
	return TRUE;
}

// Force a page swap. Try several times in case we init but can't write a bad page.
static bit EESwitch( void ){	
	idata char oldpage;
	data char i;

	oldpage = activepage;
	for( i=0; i < EEPages; i++){
		if( EEInit()){
			if( !WriteEEpromData( eebuf, next+1)) continue; // cant write: try again
			PutEEprom( oldpage | 0x80, (FLASHPTR)&ROM.pagehdr.erase);	// Make old, filled, page erase on restart
			EE_PSD_reg.PAGE = oldpage;				// Go and invalidate that page
			PutEEprom( 0, (FLASHPTR)&ROM.pagehdr.ident);	// by removing the identifier
// This bit is only needed for EESaveFlash 
// If this was a power loss event the erase will probably fall, but will be restarted on Restore. 
			if( erase < 0){
				erase = oldpage; 					// tell everyone we are in erase mode
				EEpromCmd( ERASE);				// Start erasing the page unless we already were doing one
			} else resume();
			return TRUE;
		} else break;
	}
	resume();
	return FALSE;
}

// Power loss or elective save - write current data in next available space
bit EESave( void) {
	FLASHPTR addr;
	unsigned char node, mask;
	char level;  
	
	suspend();
	addr = GetEEpromFree();
	if( !WriteEEpromData( eebuf, addr)) return EESwitch(); // If failed to write, start a new page
// OK we've written the data, now update the tree to point at it by clearing a bit
	node = *next;
	level = nextlevel;
	nextlevel = -1;			// Ensure the updated tree gets rescanned if we do another save
	if( level > 1) mask = 1 << (level - 2);		// Clear a bit in the high byte
	else if( level > 0 ) mask = 0x80; // Clear a tree bit in the low byte
	else {													// Clear a bit in the bit map
		mask = 1;
		while( (mask & node) == 0) mask <<= 1;
		if( mask != 0x40) nextlevel = 0;  		// No need to rescan if we are still on the same byte
	}	
	PutEEprom( node & ~mask, next);					// Write the amended node to flash
	resume();
	return TRUE;
}
	
bit EESaveFlash(){	
	suspend();
	return EESwitch();
}
